/*
 ķ
   Program..: _DCGRAPH*.PRG                                
   Author...:  Roger J. Donnay                             
   Notice...: (c) DONNAY Software Designs 1987-2002        
   Date.....: Oct 31, 2002                                 
   Notes....: eXPress++ Bar Graph Classes                  
 Ľ

Inspired by Bob Drakos, Kam Bhargava and Michael Rudrich

*/

#include 'dcdialog.ch'
#include 'appevent.ch'
#include 'dcgraph.ch'
#include 'dcbitmap.ch'

FUNCTION SalesGraph( nMode, nType )

LOCAL i, aCaption, aBarColor, aData, oGraph, GetList[0], bLbDown, GetOptions, ;
      oStaticText, oStaticGraph, aGraphData[0], nYear := 2000

DEFAULT nMode := 1, ;
        nType := DCGRAPH_TYPE_VERTICAL

aData := { ;
   { {10,-12,40},  ;
     {22,16,20},   ;
     {25,35,11},   ;
     {20,25,88},   ;
     {33,50,-99},  ;
     {45,22,16},   ;
     {57,90,77},   ;
     {31,-44,55},  ;
     {21,60,32},   ;
     {51,11,28},   ;
     {66,30,-19},  ;
     {62,49,73} }, ;
   { {39,-52,50},  ;
     {52,28,30},   ;
     {75,25,31},   ;
     {10,15,-28},   ;
     {63,20,29},  ;
     {35,-72,56},   ;
     {27,-30,37},   ;
     {51,14,-65},  ;
     {21,40,32},   ;
     {61,21,-18},   ;
     {46,40,29},  ;
     {32,59,23} }, ;
   { {20,-2,30},   ;
     {32,6,10},    ;
     {15,25,1},    ;
     {10,15,78},   ;
     {23,40,-89},  ;
     {35,12,6},    ;
     {47,80,67},   ;
     {21,-34,45},  ;
     {11,50,22},   ;
     {41,1,18},    ;
     {56,20,-9},   ;
     {52,39,63} } }

nYear := 2000
_GetGraphData( aGraphData, @nYear, aData )

IF nMode == 1

  oGraph := DC_BarGraph():new(,,,{600,400})
  oGraph:graphColor := { GRA_CLR_YELLOW, GRA_CLR_BLUE, GRA_CLR_RED, GRA_CLR_GREEN }
  oGraph:graphData := aGraphData
  oGraph:lbDown := {|n|DC_Winalert('This is receipt #' + aData[n,5])}
  oGraph:rbDown := oGraph:lbDown
  oGraph:textFont := '12.Arial Bold'
  oGraph:textColorFG := GRA_CLR_BLACK
  oGraph:textColorBG := nil // GRA_CLR_WHITE
  oGraph:scaleFont := '8.Arial Bold'
  oGraph:scaleColorFG := GRA_CLR_BLACK
  oGraph:type := nType
  oGraph:title := 'Year 2001 Medallions'
  oGraph:graphTitle := { 'Sales','Expenses','Profit','Eleven' }
  oGraph:graphTitleFont := '18.Arial Bold'
  oGraph:graphBackgroundColor := GRA_CLR_WHITE
  oGraph:scrollRange := 12
  oGraph:scrollButtons := .t.

  oGraph:create()

  RETURN nil

ENDIF

bLbDown := {|a,b|DC_DebugQOut({aGraphData[a,1,b]})}

aCaption := { 'Sales','Expenses','Profit' }
aBarColor := { GRA_CLR_BLUE, GRA_CLR_DARKPINK, GRA_CLR_GREEN }

@ 0,0 DCSTATIC TYPE XBPSTATIC_TYPE_TEXT ;
      SIZE 100,4 OBJECT oStaticText ;
      COLOR GRA_CLR_BLACK, GRA_CLR_WHITE

@ .5,40 DCSAY {||'Year ' + Alltrim(Str(nYear))} SAYSIZE 20,2 FONT '18.Arial Bold' ;
      COLOR GRA_CLR_BLACK, GRA_CLR_WHITE GROUP 'GRAPH'

FOR i := 1 TO Len(aCaption)
  @ 2.5,i*20 DCSAY aCaption[i] COLOR aBarColor[i] SAYSIZE 0 ;
    FONT '14.Arial Bold' PARENT oStaticText
NEXT

@ 4,0 DCSTATIC TYPE XBPSTATIC_TYPE_TEXT ;
      SIZE 100,20 OBJECT oStaticGraph ;
      COLOR GRA_CLR_BLACK, GRA_CLR_WHITE

@ 0,0 DCBARGRAPH ;
      SIZE 100,20 ;
      DATA aGraphData ;
      RBDOWN bLbDown ;
      LBDOWN bLbDown ;
      TEXTFONT '12.Arial Bold' ;
      SCALEFONT '8.Arial Bold' ;
      TEXTCOLOR GRA_CLR_BLACK ;
      SCALECOLOR GRA_CLR_BLACK ;
      BARCOLOR aBarColor ;
      TYPE nType ;
      SCROLLRANGE 6 ;
      SCROLLBUTTONS ;
      FONT '18.Arial Bold' ;
      BGCOLOR GRA_CLR_YELLOW ;
      PARENT oStaticGraph ;
      OBJECT oGraph

@ 25,0 DCPUSHBUTTON CAPTION 'Prev Year' SIZE 11,1.2 ;
      ACTION {||nYear--,_GetGraphData(aGraphData,@nYear,aData), ;
                oGraph:refresh(), DC_GetRefresh(GetList,,,,'GRAPH')} ;
      WHEN {||nYear>2000} TOOLTIP 'Graph Previous Year' ;
      GROUP 'GRAPH'

@ 25,13 DCPUSHBUTTON CAPTION 'Next Year' SIZE 11,1.2 ;
      ACTION {||nYear++,_GetGraphData(aGraphData,@nYear,aData), ;
                oGraph:refresh(), DC_GetRefresh(GetList,,,,'GRAPH')} ;
      WHEN {||nYear<2002} TOOLTIP 'Graph Next Year' ;
      GROUP 'GRAPH'

DCGETOPTIONS ;
    AUTORESIZE ;
    TOOLTIPCOLOR GRA_CLR_BLACK, GRA_CLR_CYAN ;
    TOOLTIPFONT '12.Courier'

DCREAD GUI FIT OPTIONS GetOptions TITLE 'Graph of Sales (2000-2002)'

RETURN nil

* -------------

STATIC FUNCTION _GetGraphData( aGraphData, nYear, aData )

LOCAL i, nPointer := nYear-1999, aMonths

aMonths := {'Jan','Feb','Mar','Apr','May','Jun','Jul','Aug','Sep','Oct','Nov','Dec'}
ASize(aGraphData,0)

FOR i := 1 TO 12
  AAdd(aGraphData, { aData[nPointer,i], aMonths[i] } )
NEXT

RETURN nil

* -------------

CLASS DC_BarGraph

EXPORTED:
  VAR whenBlock
  VAR hideBlock
  VAR getList
  VAR getListPointer
  VAR userCargo
  VAR optionArray
  VAR type
  VAR graphColor
  VAR graphData
  VAR textColorFG
  VAR textColorBG
  VAR textFont
  VAR scale
  VAR scaleColorFG
  VAR scaleColorBG
  VAR scaleFont
  VAR scaleMaxValue
  VAR scaleMinValue
  VAR scaleTextWidth
  VAR parent
  VAR owner
  VAR pres
  VAR visible
  VAR pos
  VAR size
  VAR lbDown
  VAR rbDown
  VAR motion
  VAR graphMaxValue
  VAR graphMinValue
  VAR graphObjects
  VAR textObjects
  VAR title
  VAR graphTitle
  VAR graphTitleFont
  VAR graphStatic
  VAR graphBackgroundColor
  VAR negOffset
  VAR cargo
  VAR maxTextWidth
  VAR barToolTips
  VAR scrollRange
  VAR scrollPointer
  VAR scrollButtons
  VAR scrollButtonArray
  VAR millions
  VAR billions
  VAR thousands
  VAR lineMarkers
  VAR status

  METHOD init, create, refresh, destroy, status

ENDCLASS

* ------------

METHOD DC_BarGraph:init( oParent, oOwner, aPos, aSize, aPres, lVisible, oGetList )

LOCAL i, xOptions, aGetList, aGetListItem, aGetOptions, bBlock

::getList := oGetList
::type := DCGRAPH_TYPE_VERTICAL

::graphObjects := {}
::textObjects := {}
::graphColor := { GRA_CLR_BLUE }
::lbDown := {||nil}
::rbDown := {||nil}
::parent := oParent
::owner := oOwner
::pos := aPos
::size := aSize
::visible := .t.
::scrollButtons := .f.
::millions := 'M'
::billions := 'B'
::thousands := 'K'
::lineMarkers := {}
::status := XBP_STAT_FAILURE

IF Valtype(::getList) == 'O'

  ::getListPointer := ::getList:getListPointer
  aGetList := ::getList:getListArray
  aGetListItem := ::getList:getListArray[::getListPointer]
  aGetOptions := ::getList:getOptions
  bBlock := DC_GetBlock(aGetListItem[bGETLIST_VAR])
  IF Valtype(bBlock) == 'B'
    ::graphData := Eval(bBlock)
  ENDIF
  ::graphTitle := aGetListItem[cGETLIST_CAPTION]

  ::type := aGetListItem[nGETLIST_SUBTYPE]
  IF Empty(::type)
    ::type := DCGRAPH_TYPE_VERTICAL
  ENDIF
  ::graphTitleFont := aGetListItem[cGETLIST_FONT]

  xOptions := aGetListItem[xGETLIST_OPTIONS]
  DEFAULT xOptions := {}
  ASize(xOptions,9)
  ::textFont := xOptions[1]
  ::scaleFont := xOptions[2]
  ::graphBackgroundColor := xOptions[3]
  IF Valtype(::graphBackgroundColor) = 'A' .AND. Len(::graphBackGroundColor) == 3
    ::graphBackgroundColor := GraMakeRGBColor(::graphBackgroundColor)
  ENDIF
  ::lbDown := xOptions[4]
  ::rbDown := xOptions[5]
  ::graphMinValue := xOptions[6]
  ::graphMaxValue := xOptions[7]
  ::scrollRange := xOptions[8]
  ::scrollButtons := xOptions[9]

  xOptions := aGetListItem[xGETLIST_OPTIONS2]
  DEFAULT xOptions := {}
  ASize(xOptions,2)
  ::textColorFG := xOptions[1]
  ::textColorBG := xOptions[2]

  xOptions := aGetListItem[xGETLIST_OPTIONS3]
  DEFAULT xOptions := {}
  ASize(xOptions,2)
  ::scaleColorFG := xOptions[1]
  ::scaleColorBG := xOptions[2]

  xOptions := aGetListItem[xGETLIST_OPTIONS4]

  IF Valtype(xOptions) == 'B'
    xOptions := Eval(xOptions)
  ENDIF

  DEFAULT xOptions := ;
     { GRA_CLR_BLUE, GRA_CLR_RED, GRA_CLR_YELLOW, GRA_CLR_GREEN, GRA_CLR_DARKPINK }

  ::graphColor := AClone(xOptions)
  FOR i := 1 TO Len(::graphColor)
    IF Valtype(::graphColor[i]) == 'A'
      ::graphColor[i] := GraMakeRGBColor(::graphColor[i])
    ENDIF
  NEXT

  bBlock := DC_GetBlock(aGetListItem[bGETLIST_GROUP],'O')
  IF Valtype(bBlock) = 'B'
    Eval( bBlock, self )
  ENDIF

  bBlock := DC_GetBlock(aGetListItem[bGETLIST_PREEVAL])
  IF Valtype(bBlock) = 'B'
    Eval( bBlock, self )
  ENDIF

  ::lineMarkers := aGetListItem[xGETLIST_OPTIONS5]
  DEFAULT ::lineMarkers := {}

ENDIF

DEFAULT ::graphMaxValue := 0, ;
        ::graphMinValue := 0

::status := XBP_STAT_INIT

RETURN self

* ------------

METHOD DC_BarGraph:create()

LOCAL GetList, GetOptions, i, j, aTextBox[0], nMaxTextWidth := 0, ;
      nMaxTextHeight := 0, lStagger, nscaleOffset, nScalePixels, ;
      nScaleTextHeight, nSpacing, nVertOffset, nRow, nCol, nBarHeight, ;
      nBarWidth, aGraphData, aGraphColor, aBarToolTips, xToolTip, ;
      aGraphTitle, nStartRow, nEndRow, aTextSize, nLastRow, cTextToolTip, ;
      cScale, nBarOffset, nTextOffset, nHeight, nWidth, nAlign, ;
      nCount, nScrollCount, oButtonStatic, nPos, aMarkerLines

IF Valtype(::getList) == 'A'
  GetList := ::getList
ELSE
  GetList := {}
  ::getList := GetList
ENDIF

DEFAULT ::pos := {}
ASize(::pos,2)

DEFAULT ::pos[1] := 0, ;
        ::pos[2] := 0

DEFAULT ::size := {}
ASize(::size,2)

DEFAULT ::size[1] := 0, ;
        ::size[2] := 0, ;
        ::scrollPointer := 1

IF Valtype(::graphData) # 'A' .OR. Empty(::graphData)
  RETURN .f.
ENDIF

// ::graphData := AClone(::graphData)

DEFAULT ::scrollRange := Len(::graphData)

IF Empty(::type)
  ::type := DCGRAPH_TYPE_VERTICAL
ENDIF

aTextBox := {}
aSize(::graphObjects,::scrollRange)
aSize(::textObjects,::scrollRange)
aSize(aTextBox,Len(::graphData))

FOR i := 1 TO Len(::graphData)
  DEFAULT ::graphData[i] := {}
  IF Len(::graphData[i]) < 6
    ASize(::graphData[i],6)
  ENDIF
  aTextBox[i] := DC_GraQueryTextBox(::graphData[i,2],::textFont)
  nMaxTextWidth := Max( nMaxTextWidth, aTextBox[i,1] )
  nMaxTextHeight := Max( nMaxTextHeight, aTextBox[i,2] )
  aGraphData := ::graphData[i,1]
  IF Valtype(aGraphData) == 'N'
    aGraphData := { aGraphData }
    ::graphData[i,1] := {::graphData[i,1]}
  ENDIF
  IF Valtype(::graphData[i,4]) # 'A'
    xToolTip := ::graphData[i,4]
    ::graphData[i,4] := Array(Len(aGraphData))
    FOR j := 1 TO Len(aGraphData)
      ::graphData[i,4,j] := xToolTip
    NEXT j
  ENDIF
  FOR j := 1 TO Len(aGraphData)
    ::graphMaxValue := Max(::graphMaxValue,aGraphData[j])
    ::graphMinValue := Min(::graphMinValue,aGraphData[j])
    IF Empty(::graphData[i,4,j])
      ::graphData[i,4,j] := _ToolTip(self,::graphData,i,j)
    ENDIF
  NEXT j
NEXT i

IF Empty(::scaleMaxValue)
  ::scaleMaxValue := ::graphMaxValue * 1.3
  ::graphMaxValue := ::scaleMaxValue
ELSE
  ::graphMaxValue := ::scaleMaxValue
ENDIF

IF Empty(::scaleMinValue)
  ::scaleMinValue := ::graphMinValue * 1.3
  ::graphMinValue := ::scaleMinValue
ELSE
  ::graphMinValue := ::scaleMinValue
ENDIF

DEFAULT ::maxTextWidth := nMaxTextWidth
DEFAULT ::scrollButtons := .f.

nScaleTextHeight := DC_GraQueryTextBox('10000',::scaleFont)[2]

IF Empty(::scale)
  ::scale := Int((::graphMaxValue-::graphMinValue)/100)
  ::scale := Val('1'+Repl('0',Len(Alltrim(Str(::scale)))))
ENDIF

IF (::graphMaxValue-::graphMinValue) <= 0
  nScalePixels := 0
ELSEIF ::type == DCGRAPH_TYPE_VERTICAL
  nScalePixels := Int(::size[2]*(::scale/(::graphMaxValue-::graphMinValue)))
ELSE
  nScalePixels := Int(::size[1]*(::scale/(::graphMaxValue-::graphMinValue)))
ENDIF

IF Empty(::scaleTextWidth)
  IF ::type == DCGRAPH_TYPE_VERTICAL
    ::scaleTextWidth := 50
  ELSE
    ::scaleTextWidth := 40
  ENDIF
ENDIF

IF ::scrollRange <= 0
  nSpacing := 0
ELSEIF ::type == DCGRAPH_TYPE_VERTICAL
  nSpacing := Int((::size[1]-::scaleTextWidth)/::scrollRange)
ELSE
  nSpacing := Int((::size[2]-10 - IIF(::scrollButtons,20,0)-::maxTextWidth)/::scrollRange)
ENDIF

lStagger := ::type = DCGRAPH_TYPE_VERTICAL .AND. ::maxTextWidth > nSpacing

IF ::scale <= 0
  ::scaleMinValue := 0
ELSE
  ::scaleMinValue := Int(::scaleMinValue/::scale) * ::scale
ENDIF
nVertOffset := nMaxTextHeight+IIF(lStagger,nMaxTextHeight,0)
nStartRow := nVertOffset + (nScaleTextHeight/2)
IF ::graphMaxValue-::graphMinValue <= 0
  ::negOffset := 0
  nEndRow := 0
ELSEIF ::type == DCGRAPH_TYPE_VERTICAL
  ::negOffset := Int(::size[2])*::scaleMinValue/(::graphMaxValue-::graphMinValue)
  nEndRow := nStartRow + ::size[2] - nVertOffset - ::negOffset
ELSE
  ::negOffset := Int(::size[1])*::scaleMinValue/(::graphMaxValue-::graphMinValue)
  nEndRow := nStartRow + ::size[1] - nVertOffset - ::negOffset
ENDIF
nVertOffset -= ::negOffset

nLastRow := -200
nCount := 1

// paint scale
FOR i := ::scaleMinValue TO ::scaleMaxValue STEP ::scale
  IF ::type == DCGRAPH_TYPE_VERTICAL
    IF ::graphMaxValue-::graphMinValue <= 0
      nRow := 0
    ELSE
      nRow := nVertOffset + Int(::size[2]*(i/(::graphMaxValue-::graphMinValue))) + 2
    ENDIF
    nCol := 0
  ELSE
    IF ::graphMaxValue-::graphMinValue <= 0
      nCol := 0
    ELSE
      nCol := nVertOffset + Int(::size[1]*(i/(::graphMaxValue-::graphMinValue))) + 2
    ENDIF
    IF nCount % 2 = 0
      nRow := 3
    ELSE
      nRow := 20
    ENDIF
    @ 0, nCol-10 DCSTATIC TYPE XBPSTATIC_TYPE_TEXT ;
      COLOR GRA_CLR_BLACK,GRA_CLR_BLACK ;
      SIZE 1, 40  ;
      PREEVAL {|o|o:clipSiblings := .t.}
  ENDIF
  nCount++
  IF i >= ::scaleMaxValue .OR. nRow > ::size[2] - IIF(::scrollButtons,20,0)
    EXIT
  ENDIF
  IF ::type == DCGRAPH_TYPE_VERTICAL
    IF nRow > nEndRow
      EXIT
    ENDIF
    IF nRow < nLastRow + nScaleTextHeight
      LOOP
    ENDIF
  ELSE
    IF nCol > nEndRow
      EXIT
    ENDIF
    IF nCol < nLastRow + nScaleTextHeight
      LOOP
    ENDIF
  ENDIF
  cScale := Alltrim(Str(i))

  IF i >= 1000000000 .OR. i <= -1000000000
    cScale := Alltrim(Str(i/1000000000,10,1)) + ::billions
  ELSEIF i >= 1000000 .OR. i <= -1000000
    cScale := Alltrim(Str(i/1000000,7,1)) + ::millions
  ELSEIF i >= 1000 .OR. i <= -1000
    cScale := Alltrim(Str(i/1000,4,1)) + ::thousands
  ENDIF

  IF ::type == DCGRAPH_TYPE_VERTICAL
    nWidth := ::scaleTextWidth
    nHeight := nScaleTextHeight
    cScale += ' -'
    nAlign := XBPSTATIC_TEXT_RIGHT + XBPSTATIC_TEXT_VCENTER
  ELSE
    nWidth := ::scaleTextWidth
    nHeight := nScaleTextHeight
    nAlign := XBPSTATIC_TEXT_CENTER
  ENDIF

  @ nRow, nCol ;
    DCSAY cScale ;
    SAYSIZE nWidth,nHeight ;
    SAYOPTION nAlign ;
    COLOR ::scaleColorFG, ::scaleColorBG ;
    FONT ::scaleFont ;
    PREEVAL {|o|o:clipSiblings := .f.}
  nLastRow := nRow
NEXT

IF ::type == DCGRAPH_TYPE_VERTICAL
  nRow := nStartRow
  nCol := ::scaleTextWidth
  nWidth := nSpacing * ::scrollRange + 20
  IF !Empty(::size)
    nHeight := ::size[2]-nRow
  ELSEIF !Empty(::parent)
    nHeight := ::parent:currentSize()[2]-nRow
  ELSE
    nHeight := nEndRow-nStartRow+::negOffset
  ENDIF
ELSE
  nCol := ::maxTextWidth + 10
  nRow := 40
  nWidth := nEndRow-nStartRow+::negOffset
  IF !Empty(::size)
    nHeight := ::size[2]-nRow
  ELSEIF !Empty(::parent)
    nHeight := ::parent:currentSize()[2]-nRow
  ELSE
    nHeight := nSpacing * ::scrollRange + 20
  ENDIF
ENDIF
IF ::scrollButtons
  nHeight -= 20
ENDIF

// Paint the STATIC for the bar graph
@ nRow, nCol DCSTATIC TYPE XBPSTATIC_TYPE_TEXT ;
  SIZE nWidth, nHeight ;
  OBJECT ::graphStatic ;
  EVAL {|o|o:setColorBG(::graphBackgroundColor)}

// Paint the scrolling buttons
IF ::scrollButtons

  @ nRow+nHeight,nCol DCSTATIC TYPE XBPSTATIC_TYPE_RAISEDBOX ;
    OBJECT oButtonStatic SIZE nWidth,20

  ::scrollButtonArray := Array(6)
  @ 0,0 DCPUSHBUTTON CAPTION BITMAP_TOP_S ;
    SIZE 20,20 ;
    OBJECT ::scrollButtonArray[1] ;
    PARENT oButtonStatic ;
    ACTION {||IIF(::scrollPointer>1, ;
                  (::scrollPointer := 1,::refresh()),nil)}

  @ 0, 20 DCPUSHBUTTON CAPTION BITMAP_REWIND_S ;
    SIZE 20,20 ;
    OBJECT ::scrollButtonArray[2] ;
    PARENT oButtonStatic ;
    ACTION {||IIF(::scrollPointer>1, ;
               (::scrollPointer -= ::scrollRange, ;
                IIF(::scrollPointer < 1, ::scrollPointer := 1, nil ), ;
                ::refresh()),nil)}

  @ 0,40 DCPUSHBUTTON CAPTION BITMAP_PREVIOUS_S ;
    SIZE 20,20 ;
    OBJECT ::scrollButtonArray[3] ;
    PARENT oButtonStatic ;
    ACTION {||IIF(::scrollPointer>1, ;
                 (::scrollPointer --, ;
                   IIF(::scrollPointer < 1, ::scrollPointer := 1, nil ), ;
                  ::refresh()),nil)}

  @ 0, nWidth-(20*4) DCPUSHBUTTON CAPTION BITMAP_NEXT_S ;
    SIZE 20,20;
    OBJECT ::scrollButtonArray[4] ;
    PARENT oButtonStatic ;
    ACTION {||IIF(::scrollPointer<Len(::graphData)-::scrollRange+1, ;
                  (::scrollPointer ++, ;
                   IIF(::scrollPointer > Len(::graphData), ::scrollPointer := Len(::graphData), nil ), ;
                   ::refresh()),nil)}

  @ 0, nWidth-(20*3) DCPUSHBUTTON CAPTION BITMAP_FORWARD_S ;
    SIZE 20,20 ;
    OBJECT ::scrollButtonArray[5] ;
    PARENT oButtonStatic ;
    ACTION {||IIF(::scrollPointer<Len(::graphData)-::scrollRange+1, ;
                  (::scrollPointer += ::scrollRange, ;
                    IIF(::scrollPointer > Len(::graphData)-::scrollRange+1, ;
                     ::scrollPointer := Len(::graphData), nil ), ;
                     ::refresh()),nil)}

  @ 0, nWidth-(20*2) DCPUSHBUTTON CAPTION BITMAP_BOTTOM_S ;
    SIZE 20,20;
    OBJECT ::scrollButtonArray[6] ;
    PARENT oButtonStatic ;
    ACTION {||::scrollPointer := Len(::graphData) - ::scrollRange + 1, ;
              ::refresh()}

ENDIF

aGraphData := ::graphData[1,1]
IF Valtype(aGraphData) == 'N'
  aGraphData := { aGraphData }
ENDIF

aGraphColor := ::graphColor
IF Valtype(aGraphColor) == 'N'
  aGraphColor := { aGraphColor }
ENDIF

IF ::type == DCGRAPH_TYPE_VERTICAL
  nBarWidth :=  (nSpacing/Len(aGraphData)) -5 + Len(aGraphData)
  nBarHeight := 10
ELSE
  nBarWidth := 10
  nBarHeight := (nSpacing/Len(aGraphData)) -5 +Len(aGraphData)
ENDIF

nScrollCount := ::scrollPointer - 1
FOR i := ::scrollPointer TO Len(::graphData)

  nScrollCount++
  IF nScrollCount > ::scrollRange
    EXIT
  ENDIF
  aGraphData := ::graphData[i,1]
  IF Valtype(aGraphData) == 'N'
    aGraphData := { aGraphData }
  ENDIF

  cTextToolTip := ::graphData[i,3]
  aBarToolTips := ::graphData[i,4]
  aMarkerLines := ::graphData[i,6]

  IF Valtype(aBarToolTips) # 'A'
    aBarToolTips := { aBarToolTips }
    ASize(aBarToolTips,Len(aGraphData))
  ENDIF

  IF Valtype(aMarkerLines) = 'A' .AND. Valtype(aMarkerLines[1]) # 'A'
    aMarkerLines := { aMarkerLines }
  ELSEIF Valtype(aMarkerLines) # 'A'
    aMarkerLines := Array(Len(aGraphData))
  ENDIF

  nBarOffset := Int((nSpacing-(nBarWidth*Len(aGraphData)))/2)
  IF ::type == DCGRAPH_TYPE_VERTICAL
    nTextOffset := Int((nSpacing-::maxTextWidth)/2)
    nRow := IIF(lStagger,(i%2)*nMaxTextHeight,0) + 3
    nCol := ::scaleTextWidth + (i-1)*nSpacing + nTextOffset
    nAlign := XBPSTATIC_TEXT_RIGHT
  ELSE
    nTextOffset := 30
    nCol := 5
    nRow := nMaxTextHeight + (i-1)*nSpacing + nTextOffset
    nAlign := XBPSTATIC_TEXT_CENTER
  ENDIF

  // Paint Text Prompts
  IF !Empty(::graphData[i,2])
    @ nRow, nCol ;
              DCSAY ::graphData[i,2] ;
              SAYSIZE ::maxTextWidth+2, nMaxTextHeight ;
              SAYOPTION nAlign ;
              SAYRIGHT ;
              FONT ::textfont ;
              COLOR ::textColorFG, ::textColorBG ;
              OBJECT ::textObjects[i] ;
              TOOLTIP cTextToolTip ;
              PARENT ::parent ;
              PREEVAL {|o|o:autoSize := .t.}
  ENDIF

  ::graphObjects[i] := Array(Len(aGraphData))

  // paint bars
  FOR j := 1 TO Len(aGraphData)

    nCol := (i-1)*nSpacing + (nBarWidth*(j-1)) + nBarOffset

    IF ::type == DCGRAPH_TYPE_VERTICAL
      nRow := -::negOffset
    ELSE
      nRow := nCol
      nCol := -::negOffset
    ENDIF

    @ nRow, nCol ;
         DCSTATIC TYPE XBPSTATIC_TYPE_TEXT ;
         COLOR aGraphColor[j], aGraphColor[j] ;
         SIZE nBarWidth,nBarHeight ;
         OBJECT ::graphObjects[i,j] PIXEL ;
         EVAL _EvalBlocks(self,i,j) ;
         TOOLTIP aBarToolTips[j] ;
         PARENT ::graphStatic

    IF Valtype(aMarkerLines[j]) = 'A'
      ASize(aMarkerLines[j],3)
      IF ::type == DCGRAPH_TYPE_VERTICAL
        @ 0,0 DCSTATIC TYPE XBPSTATIC_TYPE_TEXT ;
           PARENT ::graphObjects[i,j] PIXEL ;
           SIZE nBarWidth, 1 ;
           OBJECT aMarkerLines[j,3] ;
           COLOR aMarkerLines[j,2], aMarkerLines[j,2]
      ELSE
        @ 0,0 DCSTATIC TYPE XBPSTATIC_TYPE_TEXT ;
           PARENT ::graphObjects[i,j] PIXEL ;
           SIZE 1, nBarHeight ;
           OBJECT aMarkerLines[j,3] ;
           COLOR aMarkerLines[j,2], aMarkerLines[j,2]
      ENDIF
    ENDIF

  NEXT j

  IF Len(aGraphData) > 1
    IF ::type == DCGRAPH_TYPE_VERTICAL
      nRow := 0
      nCol := i*nSpacing
      nWidth := 1
      nHeight := nEndRow-nStartRow
    ELSE
      nCol := 0
      nRow := i*nSpacing
      nHeight := 1
      nWidth := nEndRow-nStartRow
    ENDIF

    // paint separator lines
    @ nRow, nCol DCSTATIC TYPE XBPSTATIC_TYPE_TEXT ;
         SIZE nWidth, nHeight PARENT ::graphStatic ;
         COLOR GRA_CLR_BLACK, GRA_CLR_BLACK
  ENDIF

NEXT i

IF ::type == DCGRAPH_TYPE_VERTICAL

  @ -::negOffset, 0 DCSTATIC TYPE XBPSTATIC_TYPE_TEXT ;
    COLOR GRA_CLR_BLACK,GRA_CLR_BLACK ;
    SIZE nSpacing * ::scrollRange + 20, 1 PARENT ::graphStatic

  // paint line markers
  FOR j := 1 TO Len(::lineMarkers)
    ASize(::lineMarkers[j],3)
    IF Valtype(::lineMarkers[j,2]) == 'A'
      ::lineMarkers[j,2] := GraMakeRGBColor(::lineMarkers[j,2])
    ENDIF
    DEFAULT ::lineMarkers[j,2] := GRA_CLR_BLACK
    IF ::graphMaxValue-::graphMinValue <= 0
      nPos := 0
    ELSE
      nPos := Abs(Int(::size[2]*(::lineMarkers[j,1]/(::graphMaxValue-::graphMinValue))))
    ENDIF
    @ -::negOffset + nPos, 0 DCSTATIC TYPE XBPSTATIC_TYPE_TEXT ;
      COLOR ::lineMarkers[j,2], ::lineMarkers[j,2] ;
      SIZE nSpacing * ::scrollRange + 20, 1 PARENT ::graphStatic ;
      OBJECT ::lineMarkers[j,3]
  NEXT

ELSE

  @ 0, -::negOffset DCSTATIC TYPE XBPSTATIC_TYPE_TEXT ;
    COLOR GRA_CLR_BLACK,GRA_CLR_BLACK ;
    SIZE 1, nSpacing * ::scrollRange + 20 PARENT ::graphStatic

   // paint line markers
  FOR j := 1 TO Len(::lineMarkers)
    ASize(::lineMarkers[j],3)
    DEFAULT ::lineMarkers[j,2] := GRA_CLR_BLACK
    IF ::graphMaxValue-::graphMinValue <= 0
      nPos := 0
    ELSE
      nPos := Abs(Int(::size[1]*(::lineMarkers[j,1]/(::graphMaxValue-::graphMinValue))))
    ENDIF
    @ 0, -::negOffset + nPos DCSTATIC TYPE XBPSTATIC_TYPE_TEXT ;
      COLOR ::lineMarkers[j,2], ::lineMarkers[j,2] ;
      SIZE 1, nSpacing * ::scrollRange + 20 PARENT ::graphStatic ;
      OBJECT ::lineMarkers[j,3]
  NEXT

ENDIF

aGraphData := ::graphData[1,1]
IF Valtype(aGraphData) == 'N'
  aGraphData := { aGraphData }
ENDIF

IF !Empty(::graphTitle)
  aGraphTitle := ::graphTitle
  IF Valtype(aGraphTitle) == 'C'
    aGraphTitle := { aGraphTitle }
  ENDIF
  nRow := nStartRow + nEndRow - 20 + ::negOffset
  nCol := 100
  FOR j := 1 TO Len(aGraphData)
    @ nRow,nCol DCSAY aGraphTitle[j] ;
      SAYSIZE 0 ;
      COLOR aGraphColor[j] ;
      FONT ::graphTitleFont
    nCol += DC_GraQueryTextBox( aGraphTitle[j], ::graphTitleFont )[1] + 50
  NEXT

  IF Empty(::parent)
    @ nRow+5, 0 DCPUSHBUTTON CAPTION 'Print' ;
            SIZE 50,20 ;
            ACTION {|a,b,o|o:hide(), DC_Scrn2ClipBoard(),DC_PrintClipBoard(.T.), o:show()}
  ENDIF
ENDIF

DCGETOPTIONS ;
    NOTRANSLATE ;
    PIXEL ;
    TOOLTIPCOLOR GRA_CLR_BLACK, GRA_CLR_YELLOW ;
    TOOLTIPFONT '12.Arial'

IF Valtype(::parent) == 'O'
  ::parent:lockUpdate(.t.)
ENDIF

DCREAD GUI ;
    PARENT ::parent ;
    OWNER ::owner ;
    _FIT Empty(::parent) ;
    _EXIT !Empty(::parent) ;
    OPTIONS GetOptions ;
    EVAL {|o|::refresh()} ;
    TITLE ::title ;
    SAVE ;
    _SETAPPWINDOW Empty(::parent)

IF Valtype(::parent) == 'O'
  ::parent:lockUpdate(.f.)
ENDIF

::status := XBP_STAT_CREATE

RETURN .t.

* -------------

STATIC FUNCTION _ToolTip( oGraph, aGraphData, i, j )

RETURN {||Alltrim(Transform(aGraphData[oGraph:scrollPointer+i-1,1,j],'999,999,999,999.99'))}

* -------------

STATIC FUNCTION _EvalBlocks( oGraph, i, j )

RETURN {|o|o:lbDown := {||Eval(oGraph:lbDown,oGraph:scrollPointer+i-1,j)}, ;
           o:rbDown := {||Eval(oGraph:rbDown,oGraph:scrollPointer+i-1,j)} }

* -------------

METHOD DC_BarGraph:refresh()

LOCAL i, j, nVertSize, nHorizSize, aGraphData, cText, nWidth, nHeight, ;
      nCol, nRow, nPointer, nScrollCount, aGraphColor, nGraphColor, nPos, ;
      aMarkerLines, nColor, bError, lError

IF Empty(::graphStatic)
  RETURN self
ENDIF

// ::graphData := AClone(::graphData)

lError := .t.
BEGIN SEQUENCE

// bError := ErrorBlock({||Break()})

::graphStatic:lockUpdate(.t.)

FOR i := 1 TO Len(::graphObjects)
  FOR j := 1 TO Len(::graphObjects[i])
    IF ::type == DCGRAPH_TYPE_VERTICAL
      nWidth := ::graphObjects[i,j]:currentSize()[1]
      nHeight := 0
      nCol := ::graphObjects[i,j]:currentPos()[1]
      nRow := -::negOffset
    ELSE
      nHeight := ::graphObjects[i,j]:currentSize()[2]
      nWidth := 0
      nRow := ::graphObjects[i,j]:currentPos()[2]
      nCol := -::negOffset
    ENDIF
    ::graphObjects[i,j]:setSize({nWidth,nHeight})
    ::graphObjects[i,j]:setPos({nCol,nRow})
  NEXT j
NEXT i

nPointer := ::scrollPointer-1
FOR i := 1 TO Len(::graphObjects)

  nPointer++
  IF i > ::scrollRange .OR. nPointer > Len(::graphData)
    EXIT
  ENDIF
  aGraphData := ::graphData[nPointer,1]
  cText := ::graphData[nPointer,2]
  IF Valtype(cText) == 'B'
    cText := Eval(cText)
  ENDIF
  IF Valtype(aGraphData) == 'N'
    aGraphData := { aGraphData }
  ENDIF

  aGraphColor := ::graphData[nPointer,5]
  IF Valtype(aGraphColor) $ 'NB'
    aGraphColor := { aGraphColor }
  ENDIF

  aMarkerLines := ::graphData[nPointer,6]

  IF Valtype(aMarkerLines) = 'A' .AND. Valtype(aMarkerLines[1]) # 'A'
    aMarkerLines := { aMarkerLines }
  ELSEIF Valtype(aMarkerLines) # 'A'
    aMarkerLines := Array(Len(aGraphData))
  ENDIF

  ::textObjects[i]:setsize({30,10})

  ::textObjects[i]:setCaption(Trim(DC_XtoC(cText)))

  FOR j := 1 TO Len(aGraphData)

    IF ::type == DCGRAPH_TYPE_VERTICAL
      IF ::graphMaxValue-::graphMinValue <= 0
         nVertSize := 0
      ELSE
         nVertSize := Abs(Int(::size[2]*(aGraphData[j]/(::graphMaxValue-::graphMinValue))))
      ENDIF
      nHorizSize := ::graphObjects[i,j]:currentSize()[1]
    ELSE
      IF ::graphMaxValue-::graphMinValue <= 0
        nHorizSize := 0
      ELSE
        nHorizSize := Abs(Int(::size[1]*(aGraphData[j]/(::graphMaxValue-::graphMinValue))))
      ENDIF
      nVertSize := ::graphObjects[i,j]:currentSize()[2]
    ENDIF

    IF Valtype(aGraphColor) == 'A'
      nGraphColor := aGraphColor[j]
      IF Valtype(nGraphColor) = 'B'
        nGraphColor := Eval(nGraphColor,aGraphData[j])
      ENDIF
      IF Valtype(nGraphColor) = 'A'
        nGraphColor := GraMakeRGBColor(nGraphColor)
      ENDIF
      IF Valtype(nGraphColor) = 'N'
        ::graphObjects[i,j]:setColorBG(nGraphColor)
      ENDIF
    ENDIF

    ::graphObjects[i,j]:setSize( { nHorizSize, nVertSize } )

    IF Valtype(aMarkerLines[j]) = 'A'

      IF Len(aMarkerLines[j]) < 3
        ASize(aMarkerLines[j],3)
        aMarkerLines[j,3] := ::graphObjects[i,j]:childList()[1]
      ENDIF

      IF ::type == DCGRAPH_TYPE_VERTICAL
        IF ::graphMaxValue-::graphMinValue <= 0
          nPos := 0
        ELSE
          nPos := Abs(Int(::size[2]*(aMarkerLines[j,1]/(::graphMaxValue-::graphMinValue))))
        ENDIF
        aMarkerLines[j,3]:setPos({0,nPos})
      ELSE
        IF ::graphMaxValue-::graphMinValue <= 0
          nPos := 0
        ELSE
          nPos := Abs(Int(::size[1]*(aMarkerLines[j,1]/(::graphMaxValue-::graphMinValue))))
        ENDIF
        aMarkerLines[j,3]:setPos({nPos,0})
      ENDIF
      nColor := aMarkerLines[j,2]
      IF Valtype(nColor) = 'B'
        nColor := Eval(nColor,aGraphData[j])
      ENDIF
      IF Valtype(nColor) = 'A'
        nColor := GraMakeRGBColor(nColor)
      ENDIF
      IF Valtype(nColor)='N'
        aMarkerLines[j,3]:setColorBG(nColor)
      ENDIF
    ENDIF

    IF aGraphData[j] < 0
      IF ::type == DCGRAPH_TYPE_VERTICAL
        nCol := ::graphObjects[i,j]:currentPos()[1]
        nRow := -nVertSize-::negOffset
      ELSE
        nRow := ::graphObjects[i,j]:currentPos()[2]
        nCol := -nHorizSize-::negOffset
      ENDIF
      ::graphObjects[i,j]:setPos( { nCol, nRow } )
    ENDIF

  NEXT j

NEXT i

IF Valtype(::scrollButtonArray) == 'A'
  IF ::scrollPointer == 1
    ::scrollButtonArray[1]:setCaption(BITMAP_TOP_DISABLED_S)
    ::scrollButtonArray[2]:setCaption(BITMAP_REWIND_DISABLED_S)
    ::scrollButtonArray[3]:setCaption(BITMAP_PREVIOUS_DISABLED_S)
  ELSE
    ::scrollButtonArray[1]:setCaption(BITMAP_TOP_S)
    ::scrollButtonArray[2]:setCaption(BITMAP_REWIND_S)
    ::scrollButtonArray[3]:setCaption(BITMAP_PREVIOUS_S)
  ENDIF
  IF ::scrollPointer >= Len(::graphData) - ::scrollRange + 1
    ::scrollButtonArray[6]:setCaption(BITMAP_BOTTOM_DISABLED_S)
    ::scrollButtonArray[5]:setCaption(BITMAP_FORWARD_DISABLED_S)
    ::scrollButtonArray[4]:setCaption(BITMAP_NEXT_DISABLED_S)
  ELSE
    ::scrollButtonArray[6]:setCaption(BITMAP_BOTTOM_S)
    ::scrollButtonArray[5]:setCaption(BITMAP_FORWARD_S)
    ::scrollButtonArray[4]:setCaption(BITMAP_NEXT_S)
  ENDIF
ENDIF

::graphStatic:lockUpdate(.f.)
::graphStatic:invalidateRect()

FOR i := 1 TO Len(::lineMarkers)
  IF Len(::lineMarkers[i]) >= 3
    ::lineMarkers[i,3]:toFront()
  ENDIF
NEXT

lError := .f.

END SEQUENCE
// ErrorBlock(bError)

IF lError
  ::graphStatic:hide()
ELSE
  ::graphStatic:lockUpdate(.f.)
  ::graphStatic:invalidateRect()
  ::graphStatic:show()
ENDIF

RETURN self

* -------------

METHOD DC_BarGraph:destroy()

DC_GetDestroy(::getList)

::whenBlock := nil
::hideBlock := nil
::getList := nil
::getListPointer := nil
::userCargo := nil
::optionArray := nil
::type := nil
::graphColor := nil
::graphData := nil
::textColorFG := nil
::textColorBG := nil
::textFont := nil
::scale := nil
::scaleColorFG := nil
::scaleColorBG := nil
::scaleFont := nil
::scaleMaxValue := nil
::scaleMinValue := nil
::scaleTextWidth := nil
::parent := nil
::owner := nil
::pres := nil
::visible := nil
::pos := nil
::size := nil
::lbDown := nil
::rbDown := nil
::motion := nil
::graphMaxValue := nil
::graphMinValue := nil
::graphObjects := nil
::textObjects := nil
::title := nil
::graphTitle := nil
::graphTitleFont := nil
::graphStatic := nil
::graphBackgroundColor := nil
::negOffset := nil
::cargo := nil
::maxTextWidth := nil
::barToolTips := nil
::scrollRange := nil
::scrollPointer := nil
::scrollButtons := nil
::scrollButtonArray := nil
::millions := nil
::billions := nil
::thousands := nil
::lineMarkers := nil

RETURN self

* -------------

METHOD DC_BarGraph:status()

RETURN ::status


CLASS DC_BarChart FROM DC_BarGraph

ENDCLASS
